#include "equadprog2.hpp"
#include "echolesky_decomposition.hpp"
#include "nonlinear_optimization.h"

#include <iostream>

#include "eigen_common.h"	
#include <Eigen/Cholesky>

using namespace Eigen;

NonlinearMPC::NonlinearMPC(int _N_TIMESTEPS, double _TIMESTEP){
	is_initialized=false;
	is_model_set=false;
	is_allocated=false;
	N_TIMESTEPS = _N_TIMESTEPS;
	TIMESTEP = _TIMESTEP;
}

void NonlinearMPC::set_model(mpcModel *m){
	model = m;
	is_model_set = true;
	if (TIMESTEP!=m->TIMESTEP)	printf("WARNING: Model and MPC Timesteps not equal (changing model T...)\n");
	model->TIMESTEP = TIMESTEP;
	allocate();
}

NonlinearMPC::~NonlinearMPC(){
	//deallocate();
}

void NonlinearMPC::allocate(){
	int i,j;
	
	X = (Vectord*)malloc((N_TIMESTEPS+1)*sizeof(Vectord));	// extra timestep is for final X
	U = (Vectord*)malloc(N_TIMESTEPS*sizeof(Vectord));
	Xhat = (Vectord*)malloc((N_TIMESTEPS+1)*sizeof(Vectord));	// extra timestep is for final X
	Uhat = (Vectord*)malloc(N_TIMESTEPS*sizeof(Vectord));

	K = (Matrixd*)malloc(N_TIMESTEPS*sizeof(Matrixd));
	dU = (Vectord*)malloc(N_TIMESTEPS*sizeof(Vectord));

	V = (double*)malloc((N_TIMESTEPS+1));	// extra timestep is for final X
	Vx = (Vectord*)malloc((N_TIMESTEPS+1)*sizeof(Vectord));	// extra timestep is for final X
	Vxx = (Matrixd*)malloc((N_TIMESTEPS+1)*sizeof(Matrixd));	// extra timestep is for final X

	Fx = (Matrixd*)malloc(N_TIMESTEPS*sizeof(Matrixd));
	Fu = (Matrixd*)malloc(N_TIMESTEPS*sizeof(Matrixd));
	Fxx = (Matrixd3D*)malloc(N_TIMESTEPS*sizeof(Matrixd3D));
	Fux = (Matrixd3D*)malloc(N_TIMESTEPS*sizeof(Matrixd3D));
	Fuu = (Matrixd3D*)malloc(N_TIMESTEPS*sizeof(Matrixd3D));
	
	Lx = (Vectord*)malloc(N_TIMESTEPS*sizeof(Vectord));	
	Lu = (Vectord*)malloc(N_TIMESTEPS*sizeof(Vectord));
	Lxx = (Matrixd*)malloc(N_TIMESTEPS*sizeof(Matrixd));
	Lux = (Matrixd*)malloc(N_TIMESTEPS*sizeof(Matrixd));
	Luu = (Matrixd*)malloc(N_TIMESTEPS*sizeof(Matrixd));

	Qx = (Vectord*)malloc(N_TIMESTEPS*sizeof(Vectord));	
	Qu = (Vectord*)malloc(N_TIMESTEPS*sizeof(Vectord));
	Qxx = (Matrixd*)malloc(N_TIMESTEPS*sizeof(Matrixd));
	Qux = (Matrixd*)malloc(N_TIMESTEPS*sizeof(Matrixd));
	Quu = (Matrixd*)malloc(N_TIMESTEPS*sizeof(Matrixd));
	
	for(i=0;i<N_TIMESTEPS;i++){
		X[i].setZero(model->num_states);
		U[i].setZero(model->num_actions);	
		Xhat[i].setZero(model->num_states);
		Uhat[i].setZero(model->num_actions);	
		dU[i].setZero(model->num_actions);	
		
		K[i].setZero(model->num_actions,model->num_states);

		V[i] = 0;
		Vx[i].setZero(model->num_states);
		Vxx[i].setZero(model->num_states,model->num_states);

		Fx[i].setZero(model->num_states,model->num_states);
		Fu[i].setZero(model->num_states,model->num_actions);

		Fxx[i].setZero(model->num_states,model->num_states*model->num_states);
		Fux[i].setZero(model->num_actions,model->num_states*model->num_states);
		Fuu[i].setZero(model->num_actions,model->num_actions*model->num_states);

		Lx[i].setZero(model->num_states);
		Lu[i].setZero(model->num_actions);

		Lxx[i].setZero(model->num_states,model->num_states);
		Lux[i].setZero(model->num_actions,model->num_states);
		Luu[i].setZero(model->num_actions,model->num_actions);

		Qx[i].setZero(model->num_states);
		Qu[i].setZero(model->num_actions);

		Qxx[i].setZero(model->num_states,model->num_states);
		Qux[i].setZero(model->num_actions,model->num_states);
		Quu[i].setZero(model->num_actions,model->num_actions);
	}

	V[N_TIMESTEPS] = 0;
	Vx[N_TIMESTEPS].setZero(model->num_states);
	Vxx[N_TIMESTEPS].setZero(model->num_states,model->num_states);
	X[N_TIMESTEPS].setZero(model->num_states);

	is_allocated=true;

}

void NonlinearMPC::run(double t, Vectord &x0){
	int i;

	// initialize? set gains to zero, initial trajectory, etc.
	mu_regularization = 1.0e-6;

	is_first_iter = true;
	
	for(i=0;i<5;i++){

		// computes derivatives along the nominal trajectory
		forward(t,x0);

		// propogates the value function backwards along the trajectory
		backward();

		// searches for next nominal trajectory
		forward2();

		is_first_iter = false;

		if (fabs(cost_expected_change)<1e-4){
			printf("Optimization Converged\n");
			break;
		}
	}
}

void NonlinearMPC::forward(double t, Vectord &x0){
	int i;
		
	if(!is_model_set)	printf("ERROR: Model Not Set\n");
	if(!is_allocated)	printf("ERROR: Memory Not Allocated\n");
	
	Vectord x_d(model->num_states);
	Vectord x(model->num_states);

	x = x0;

	for(i=0;i<N_TIMESTEPS;i++){
		if(is_first_iter){
			x_d = X[i];
			X[i] = x;
			U[i] += K[i]*(x - x_d);
			model->forward_dynamics(i*TIMESTEP,X[i],U[i],x);
		}else{
			// nominal trajectory already computed and stored in X,U
		}

		model->user_forward_dynamics_derivatives(i*TIMESTEP,X[i],U[i],Fx[i],Fu[i],Fxx[i],Fux[i],Fuu[i]);
		model->user_one_step_cost_derivatives(i*TIMESTEP,X[i],U[i],Lx[i],Lu[i],Lxx[i],Lux[i],Luu[i]);
		
		printf("%i. % f % f % f % f % f % f\n",i,X[i](0),X[i](1),U[i](0),dU[i](0),K[i](0,0),K[i](0,1));
	}

	X[N_TIMESTEPS] = x;
}

void NonlinearMPC::backward(){

	int i,j;
	
	// VN = final_cost(XN)
	model->user_final_cost(N_TIMESTEPS*TIMESTEP,X[N_TIMESTEPS],U[N_TIMESTEPS],Vx[N_TIMESTEPS],Vxx[N_TIMESTEPS]);

	MatrixXd Ixx = MatrixXd::Identity(model->num_states,model->num_states);

	for(i=N_TIMESTEPS-1;i>=0;i--){

		// Qx = Lx + Fx'*Vx(i+1)
		Qx[i] = Lx[i] + Fx[i].transpose()*Vx[i+1];
		
		// Qu = Lu + Fu'*Vx(i+1)
		Qu[i] = Lu[i] + Fu[i].transpose()*Vx[i+1];

		//Qxx = Lxx + Fx'*Vxx(i+1)*Fx
		Qxx[i] = Lxx[i] + Fx[i].transpose()*Vxx[i+1]*Fx[i];

		// Qux = Lux + Fu'*(Vxx(i+1) + mu*I)*Fx
		Qux[i] = Lux[i] + Fu[i].transpose()*(Vxx[i+1] + mu_regularization*Ixx)*Fx[i];

		// Quu = Luu + Fu'*(Vxx(i+1) + mu*I)*Fu
		Quu[i] = Luu[i] + Fu[i].transpose()*(Vxx[i+1] + mu_regularization*Ixx)*Fu[i];

		if (!model->is_linear){
			// loop to compute the 3D matrices
			for(j=0;j<model->num_states;j++){

				// Qxx += Vx(i+1)*Fxx	
				Qxx[i] += Vx[i+1](j)*Fxx[i].block(0,j*model->num_states,model->num_states,model->num_states);

				// Qux += Vx(i+1)*Fux
				Qux[i] += Vx[i+1](j)*Fux[i].block(0,j*model->num_states,model->num_actions,model->num_states);

				// Quu += Vx(i+1)*Fuu
				Quu[i] += Vx[i+1](j)*Fuu[i].block(0,j*model->num_states,model->num_actions,model->num_actions);
			}
		}
	
		//ColPivHouseholderQR<MatrixXd> LQ(Quu[i]);
		//if(!LQ.isInvertible())	printf("WARNING: Quu is not invertible\n");
		LDLT<MatrixXd> Quuinv(Quu[i]);

		/*************************************************/
		/* Should be doing QP here to handle constraints */
		/*
		/* dU = argmin Q(dX,dU) = 0.5*[1,dX,dU]*[0,Qx',Qu';
		/*                                       Qx,Qxx,Qxu;
		/*                                       Qu,Qux,Quu]*[1;dX;dU];
		/* such that Umin <= U + dU <= Umax
		/*************************************************/
		if(model->is_control_inequality_constrained){
			VectorXd ci0(2);
			MatrixXd CI(1,2);
			// evaluate inequality control constraints
			//model->user_control_inequality_constraints(i*TIMESTEP,X[i],U[i],c,C);
			//CI^T x + ci0 >= 0
			CI(0) = -1; ci0(0) = 0.1 - U[i](0);
			CI(1) =  1;	ci0(1) = 0.1 + U[i](0);

			VectorXd ce0;
			MatrixXd CE;

			MatrixXd G(model->num_actions,model->num_actions);
			VectorXd g0(model->num_actions);
			G = Quu[i];
			g0 = Qu[i];
			echolesky_decomposition(G);
			double Gtrace = G.diagonal().sum();

			VectorXd dUtemp(1);

			solve_quadprog2( G,  Gtrace, g0, CE, ce0, CI, ci0, dUtemp );

			dU[i](0) = dUtemp(0);

			K[i] = -Quuinv.solve(Qux[i]);

		}else{
			// k = -Quu^-1*Qu;
			dU[i] = -Quuinv.solve(Qu[i]);
			
			// K = -Quu^-1*Qux;
			K[i] = -Quuinv.solve(Qux[i]);
		}
		
		/**************************************************/
		/**************************************************/

		// V += 0.5*k'*Quu*k + k'*Qu
		V[i] += (0.5*dU[i].transpose()*Quu[i]*dU[i] + dU[i].transpose()*Qu[i]).sum();	

		// Vx = Qx + K'*Quu*k + K'*Qu + Qux'*k
		Vx[i] = Qx[i] + K[i].transpose()*Quu[i]*dU[i] + K[i].transpose()*Qu[i] + Qux[i].transpose()*dU[i];

		// Vxx = Qxx + K'*Quu*K + K'*Qux + Qux'*K
		Vxx[i] = Qxx[i] + K[i].transpose()*Quu[i]*K[i] + K[i].transpose()*Qux[i] + Qux[i].transpose()*K[i];

	}

}

void NonlinearMPC::forward2(){

	int MAX_LINE_SEARCH_ITERS = 10;

	double alpha=1;	// start with 1 and move down

	int i,j;
	Vectord x(model->num_states);

	// iteratively update alpha to determine optimal new trajectory
	for(i=0;i<MAX_LINE_SEARCH_ITERS;i++){
		
		cost_temporary = 0;
		cost_nominal = 0;
		cost_expected_change = 0;		
		x = X[0];

		// integrate and compute
		for(j=0;j<N_TIMESTEPS;j++){
			Xhat[j] = x;

			// new updated control trajectory
			Uhat[j] = U[j] + alpha*dU[j] + K[j]*(Xhat[j] - X[j]);

			// integrate new trajectory
			model->forward_dynamics(i*TIMESTEP,Xhat[j],Uhat[j],x);

			// cost of last nominal trajectory
			cost_nominal += model->one_step_cost(i*TIMESTEP,X[j],U[j]);

			// cost of new nominal trajectory
			cost_temporary += model->one_step_cost(i*TIMESTEP,Xhat[j],Uhat[j]);

			// estimated cost change
			cost_expected_change += (alpha*dU[j].transpose()*Qu[j] + alpha*alpha*0.5*dU[j].transpose()*Quu[j]*dU[j]).sum();
		}
		Xhat[N_TIMESTEPS] = x;

		printf("%2i: % f % f % f\n",i,cost_nominal,cost_temporary,cost_expected_change);
		
		if(cost_temporary<cost_nominal || fabs(cost_nominal-cost_temporary)<1e-3){
			
			// save the new trajectory as the nominal
			for(j=0;j<N_TIMESTEPS;j++){
				U[j] = Uhat[j];
				X[j] = Xhat[j];
			}
			X[N_TIMESTEPS] = Xhat[N_TIMESTEPS];
			break;	
		}

		alpha = 1.0 - (double)i/(MAX_LINE_SEARCH_ITERS);

		if(alpha<0)	alpha = 0;
		if(alpha>1) alpha = 1;

		printf("alpha = % f\n",alpha);
	}

}